# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""This module aims to provide tools to chaining of calls.

It is inspired from method chaining pattern in embedding objects to chain
methods calls in a dedicated Chaining object. Such method calls return the
Chaining object itself, allowing multiple calls to object methods to be invoked
in a concise statement.
"""

__all__ = ['Chaining', 'ListChaining']


class Chaining(object):
    """Class which permits to process chaining of routines/attr such as call
    chaining in javascript.

    In order to chain calls, start to embed an object, then call embedded
    object methods which returns all the Chaining object. Finally, if you want
    to get back method results, use Chaining.__getitem__ method where the index
    corresponds to the order of calls.

    :Example:

    >>> chaining = Chaining('example').upper().capitalize()
    >>> chaining[0]
    'EXAMPLE'
    >>> chaining[1]
    'Example'
    >>> chaining[-1]
    'example.'
    >>> chaining[:]
    ['EXAMPLE', 'Example', 'example.']
    >>> chaining._
    'example'
    >>> chaining.__iadd__('.').__iadd__('!')
    >>> chaining._
    'example.!'
    >>> chaining[:]
    ['EXAMPLE', 'Example', 'example.', None, None]
    """

    CONTENT = '_'  #: content attribute name
    RESULTS = '___'  #: chained method results attribute name

    __slots__ = (CONTENT, RESULTS)

    def __init__(self, content):

        super(Chaining, self).__init__()

        self._ = content
        self.___ = []

    def __getitem__(self, key):

        return self.___[key]

    def __getattribute__(self, key):

        if key in Chaining.__slots__:
            # check if key is in self slots
            result = super(Chaining, self).__getattribute__(key)

        else:  # else try to get key from self.content
            attr = getattr(self._, key)
            # embed routine in self._processing_name method
            result = _process_function(self, attr)

        return result


def _process_function(chaining, routine):
    """Chain function which returns a function.

    :param routine: routine to process.
    :return: routine embedding execution function.
    """

    def processing(*args, **kwargs):
        """Execute routine with input args and kwargs and add reuslt in
        chaining.___.

        :param tuple args: routine varargs.
        :param dict kwargs: routine kwargs.
        :return: chaining chaining.
        :rtype: Chaining
        """

        result = routine(*args, **kwargs)
        chaining.___.append(result)

        return chaining

    return processing


class ListChaining(Chaining):
    """Apply chaining on a list of objects.

    According to content length, chaining results are saved in a list where
        values are call result or exception if an exeception occured.

    :Example:

    >>> chaining = ListChaining('example', 'test').upper().capitalize()
    >>> chaining[0]
    ['EXAMPLE', 'TEST']
    >>> chaining[1]
    ['Example', 'Test']
    >>> chaining[-1]
    ['Example', 'Test']
    >>> chaining._
    ['example', 'test']
    >>> chaining += '.'
    >>> chaining._
    ['example.', 'test.']
    >>> chaining[2]
    [None, None]
    >>> chaining[:]
    [['EXAMPLE', 'TEST'], ['Example', 'Test'], [None, None]]
    """

    __slots__ = Chaining.__slots__

    def __init__(self, *content):

        super(ListChaining, self).__init__(content=content)

    def __getattribute__(self, key):

        if key in ListChaining.__slots__:
            result = super(ListChaining, self).__getattribute__(key)

        else:
            # list of routines to execute
            self_content = self._
            routines = [None] * len(self_content)
            # get routines from self.content and input key
            for index, content in enumerate(self_content):
                routine = None
                try:
                    routine = getattr(content, key)
                except AttributeError as excp:
                    # in case of exception, routine is the exception
                    routine = excp
                # in all cases, put routine in routines
                routines[index] = routine
            result = _process_function_list(self, routines)

        return result


def _process_function_list(self, routines):
    """Chain function which returns a function.

    :param routines: routines to process.
    """

    def processing(*args, **kwargs):
        """Execute routines with input args and kwargs and add result in
        chaining.___.

        :param tuple args: routines varargs.
        :param dict kwargs: routines kwargs.
        :return: chaining chaining.
        :rtype: Chaining
        """

        results = [None] * len(routines)
        for index, routine in enumerate(routines):
            if isinstance(routine, Exception):
                result = routine
            else:
                try:
                    result = routine(*args, **kwargs)
                except AttributeError as excp:
                    result = excp
            results[index] = result
        self.___.append(results)

        return self

    return processing



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from inspect import getargspec, isbuiltin, getmembers, isroutine

from b3j0f.utils.ut import UTCase
from b3j0f.utils.proxy import (
    get_proxy, proxify_routine, proxify_elt, proxified_elt, is_proxy
)


class ProxyRoutineTest(UTCase):
    """Test proxy routine function.
    """

    def _assert_routine(self, routine, _get_proxy=True):
        """Assert properties on a proxified routine.

        :param routine: routine to proxify.
        :param bool get_proxy: (private) play get_routine if True.
        """
        # check is not proxy
        isproxy = is_proxy(routine)
        self.assertFalse(isproxy)

        # get proxy
        proxy = get_proxy(routine) if _get_proxy else proxify_routine(routine)
        # check is proxy
        isproxy = is_proxy(proxy)
        self.assertTrue(isproxy)

        try:
            func_argspec = getargspec(routine)
        except TypeError:
            pass
        else:
            proxy_argspec = getargspec(proxy)
            self.assertEqual(func_argspec, proxy_argspec)

        if not isbuiltin(routine):
            self.assertIs(routine.__class__, proxy.__class__)
            self.assertIsNot(proxy.__dict__, routine.__dict__)
        self.assertEqual(proxy.__name__, routine.__name__)
        self.assertEqual(proxy.__doc__, routine.__doc__)
        self.assertEqual(proxy.__module__, proxify_routine.__module__)
        # assert proxified element is routine
        proxified = proxified_elt(proxy)
        self.assertIs(proxified, routine)
        # repeat assertion with proxify_routine function
        if _get_proxy:
            self._assert_routine(routine, _get_proxy=False)

    def test_function_empty(self):

        def test(a, b=1, *args, **kwargs):
            """Default test function.
            """
            pass

        self._assert_routine(test)

    def test_builtin(self):
        """Test to proxify a builtin function
        """

        self._assert_routine(min, True)

    def test_lambda(self):
        """Test lambda expression.
        """

        self._assert_routine(lambda a, b=2, *args, **kwargs: None)

    def test_method(self):
        """Test to proxify a method.
        """

        self._assert_routine(self.test_method)


class ProxyEltTest(UTCase):
    """Test proxy elt function.
    """

    def _assert_elt(self, add_bases=False, add_dict=False, public=False):
        """Assert to proxify an elt.
        """

        class A(object):
            def _protected(self):
                pass

            def a(self, a):
                return self

        class B:
            def b(self, b):
                pass

        class C(A, B):
            def test(self, test):
                pass

        elt = C()

        bases = (A,) if add_bases else ()
        _dict = {'test': lambda: None} if add_dict else None

        proxy = proxify_elt(elt, bases=bases, _dict=_dict, public=public)

        if add_bases:
            self.assertIs(proxy.a(a=None), elt)
            # check if forgiven bases are not proxified
            self.assertNotIsInstance(proxy, B)
            # check if given bases are proxified
            for base in bases:
                self.assertIsInstance(proxy, base)
                for name, _ in getmembers(base, lambda m: isroutine(m)):
                    elt_member = getattr(elt, name, None)
                    if hasattr(elt_member, '__func__'):
                        elt_member = elt_member.__func__
                        proxy_member = getattr(proxy, name).__func__
                        proxified_member = proxified_elt(proxy_member)
                        proxified_member = getattr(
                            proxified_member, '__func__', proxified_member
                        )
                        self.assertIs(proxified_member, elt_member)
                        if not isbuiltin(elt_member):
                            # compare argspec
                            elt_argspec = getargspec(elt_member)
                            proxy_argspec = getargspec(proxy_member)
                            self.assertEqual(elt_argspec, proxy_argspec)
                            self.assertIs(
                                elt_member.__class__, proxy_member.__class__
                            )
                            self.assertIsNot(
                                elt_member.__dict__, proxy_member.__dict__
                            )
                        self.assertEqual(
                            elt_member.__name__, proxy_member.__name__
                        )
                        self.assertEqual(
                            elt_member.__doc__, proxy_member.__doc__
                        )
                        self.assertEqual(
                            proxy_member.__module__, proxify_elt.__module__
                        )

        if add_dict:
            for name in _dict:
                member = _dict[name]
                proxy_member = getattr(proxy, name)
                proxified_member = proxified_elt(proxy_member)
                proxified_member = getattr(
                    proxified_member, '__func__', proxified_member
                )
                self.assertIs(proxified_member, member)

        # check public members
        if public:
            self.assertFalse(hasattr(proxy, '_protected'))
        else:
            self.assertTrue(not bases or hasattr(proxy, '_protected'))

    def test_elt(self):
        """Test to proxify an elt.
        """

        self._assert_elt()

    def test_elt_public(self):
        """Test to proxify an elt with only public members.
        """

        self._assert_elt(public=True)

    def test_elt_bases(self):
        """Test to proxify an elt with bases.
        """

        self._assert_elt(add_bases=True)

    def test_elt_dict(self):
        """Test to proxify an elt with _dict.
        """

        self._assert_elt(add_dict=True)

    def test_elt_bases_dict(self):
        """Test to proxify an elt with bases and _dict.
        """

        self._assert_elt(add_bases=True, add_dict=True)

    def test__getproxy__(self):
        """Test the __getproxy__ instance method.
        """

        testproxy = 'test'

        class Test(object):
            """Test class.
            """
            def __getproxy__(self):
                """Specific get_proxy function to use in order to proxify a
                Test instance.
                """

                return testproxy

        test = Test()

        proxy = get_proxy(elt=test)

        self.assertEqual(proxy, testproxy)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Code from http://code.activestate.com/recipes/277940-decorator-for-\
bindingconstants-at-compile-time/

Decorator for automatic code optimization.
If a global is known at compile time, replace it with a constant.
Fold tuples of constants into a single constant.
Fold constant attribute lookups into a single constant.

Modifications:

- Add constants values from opmap constants (STORE_GLOBAL, etc.) in order to
    avoid to update globals.
- Modify verbose argument which is None or use a function with one argument
    which can be bound to a print function or a logging function.
- Set attributes from originary function such as __dict__, __module__, etc.
"""

from opcode import opmap, HAVE_ARGUMENT, EXTENDED_ARG

from types import FunctionType, ModuleType

try:
    import __builtin__
except ImportError:
    import builtins as __builtin__

from b3j0f.utils.version import PY3

__all__ = ['bind_all', 'make_constants']

STORE_GLOBAL = opmap['STORE_GLOBAL']
LOAD_GLOBAL = opmap['LOAD_GLOBAL']
LOAD_CONST = opmap['LOAD_CONST']
LOAD_ATTR = opmap['LOAD_ATTR']
BUILD_TUPLE = opmap['BUILD_TUPLE']
JUMP_FORWARD = opmap['JUMP_FORWARD']

WRAPPER_ASSIGNMENTS = ('__doc__', '__annotations__', '__dict__', '__module__')


def _make_constants(f, builtin_only=False, stoplist=None, verbose=None):
    """Generate new function where code is an input function code with all
    LOAD_GLOBAL statements changed to LOAD_CONST statements.

    :param function f: code function to transform.
    :param bool builtin_only: only transform builtin objects.
    :param list stoplist: attribute names to not transform.
    :param function verbose: logger function which takes in parameter a message

    .. warning::
        Be sure global attributes to transform are not resolved dynamically.
    """

    result = f

    if stoplist is None:
        stoplist = []

    try:
        fcode = f.__code__
    except AttributeError:
        return f        # Jython doesn't have a __code__ attribute.
    newcode = list(fcode.co_code) if PY3 else [ord(co) for co in fcode.co_code]
    newconsts = list(fcode.co_consts)
    names = fcode.co_names
    codelen = len(newcode)

    env = vars(__builtin__).copy()
    if builtin_only:
        stoplist = dict.fromkeys(stoplist)
        stoplist.update(f.__globals__)
    else:
        env.update(f.__globals__)

    # First pass converts global lookups into constants
    changed = False
    i = 0
    while i < codelen:
        opcode = newcode[i]
        if opcode in (EXTENDED_ARG, STORE_GLOBAL):
            return f    # for simplicity, only optimize common cases
        if opcode == LOAD_GLOBAL:
            oparg = newcode[i + 1] + (newcode[i + 2] << 8)
            name = fcode.co_names[oparg]
            if name in env and name not in stoplist:
                value = env[name]
                for pos, val in enumerate(newconsts):
                    if val is value:
                        break
                else:
                    pos = len(newconsts)
                    newconsts.append(value)
                newcode[i] = LOAD_CONST
                newcode[i + 1] = pos & 0xFF
                newcode[i + 2] = pos >> 8
                changed = True
                if verbose is not None:
                    verbose("{0} --> {1}".format(name, value))
        i += 1
        if opcode >= HAVE_ARGUMENT:
            i += 2

    # Second pass folds tuples of constants and constant attribute lookups
    i = 0
    while i < codelen:

        newtuple = []
        while newcode[i] == LOAD_CONST:
            oparg = newcode[i + 1] + (newcode[i + 2] << 8)
            newtuple.append(newconsts[oparg])
            i += 3

        opcode = newcode[i]
        if not newtuple:
            i += 1
            if opcode >= HAVE_ARGUMENT:
                i += 2
            continue

        if opcode == LOAD_ATTR:
            obj = newtuple[-1]
            oparg = newcode[i + 1] + (newcode[i + 2] << 8)
            name = names[oparg]
            try:
                value = getattr(obj, name)
            except AttributeError:
                continue
            deletions = 1

        elif opcode == BUILD_TUPLE:
            oparg = newcode[i + 1] + (newcode[i + 2] << 8)
            if oparg != len(newtuple):
                continue
            deletions = len(newtuple)
            value = tuple(newtuple)

        else:
            continue

        reljump = deletions * 3
        newcode[i - reljump] = JUMP_FORWARD
        newcode[i - reljump + 1] = (reljump - 3) & 0xFF
        newcode[i - reljump + 2] = (reljump - 3) >> 8

        nclen = len(newconsts)
        newconsts.append(value)
        newcode[i] = LOAD_CONST
        newcode[i + 1] = nclen & 0xFF
        newcode[i + 2] = nclen >> 8
        i += 3
        changed = True
        if verbose is not None:
            verbose("new folded constant:{0}".format(value))

    if changed:

        codestr = bytes(newcode) if PY3 else ''.join(
            [chr(co) for co in newcode]
        )
        vargs = [
            fcode.co_argcount, fcode.co_nlocals, fcode.co_stacksize,
            fcode.co_flags, codestr, tuple(newconsts), fcode.co_names,
            fcode.co_varnames, fcode.co_filename, fcode.co_name,
            fcode.co_firstlineno, fcode.co_lnotab, fcode.co_freevars,
            fcode.co_cellvars
        ]
        if PY3:
            vargs.insert(1, fcode.co_kwonlyargcount)

        codeobj = type(fcode)(*vargs)
        result = type(f)(
            codeobj, f.__globals__, f.__name__, f.__defaults__, f.__closure__
        )

        # set f attributes to result
        for prop in WRAPPER_ASSIGNMENTS:
            try:
                attr = getattr(f, prop)
            except AttributeError:
                pass
            else:
                setattr(result, prop, attr)

    return result

_make_constants = _make_constants(_make_constants)  # optimize thyself!


def bind_all(morc, builtin_only=False, stoplist=None, verbose=None):
    """Recursively apply constant binding to functions in a module or class.

    Use as the last line of the module (after everything is defined, but
    before test code). In modules that need modifiable globals, set
    builtin_only to True.

    :param morc: module or class to transform.
    :param bool builtin_only: only transform builtin objects.
    :param list stoplist: attribute names to not transform.
    :param function verbose: logger function which takes in parameter a message
    """

    if stoplist is None:
        stoplist = []

    def _bind_all(morc, builtin_only=False, stoplist=None, verbose=False):
        """Internal bind all decorator function.
        """
        if stoplist is None:
            stoplist = []

        if isinstance(morc, (ModuleType, type)):
            for k, val in list(vars(morc).items()):
                if isinstance(val, FunctionType):
                    newv = _make_constants(
                        val, builtin_only, stoplist, verbose
                    )
                    setattr(morc, k, newv)
                elif isinstance(val, type):
                    _bind_all(val, builtin_only, stoplist, verbose)

    if isinstance(morc, dict):  # allow: bind_all(globals())
        for k, val in list(morc.items()):
            if isinstance(val, FunctionType):
                newv = _make_constants(val, builtin_only, stoplist, verbose)
                morc[k] = newv
            elif isinstance(val, type):
                _bind_all(val, builtin_only, stoplist, verbose)
    else:
        _bind_all(morc, builtin_only, stoplist, verbose)


@_make_constants
def make_constants(builtin_only=False, stoplist=None, verbose=None):
    """Return a decorator for optimizing global references.

    Replaces global references with their currently defined values.
    If not defined, the dynamic (runtime) global lookup is left undisturbed.
    If builtin_only is True, then only builtins are optimized.
    Variable names in the stoplist are also left undisturbed.
    Also, folds constant attr lookups and tuples of constants.
    If verbose is True, prints each substitution as is occurs.

    :param bool builtin_only: only transform builtin objects.
    :param list stoplist: attribute names to not transform.
    :param function verbose: logger function which takes in parameter a message
    """

    if stoplist is None:
        stoplist = []

    if isinstance(builtin_only, type(make_constants)):
        raise ValueError("The bind_constants decorator must have arguments.")
    return lambda f: _make_constants(f, builtin_only, stoplist, verbose)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Library which aims to bind named properties on any element at runtime.

This module can bind a named property on any element but None methods.

In preserving binding from inheritance and automatical mechanisms which prevent
to set any attribute on any elements.

When you are looking for an element bound property, the result is a dictionary
of the shape {element, {property name, property}}.

.. warning:

    - It is adviced to delete properties from cache after deleting them at
        runtime in order to avoid memory leak.
"""

__all__ = [
    'get_properties',
    'get_first_property', 'get_first_properties',
    'get_local_property', 'get_local_properties',
    'put_properties', 'del_properties',
    'firsts', 'remove_ctx',
    'setdefault', 'free_cache',
    'find_ctx', 'addproperties'
]

from b3j0f.utils.version import PY2
from b3j0f.utils.iterable import ensureiterable

from inspect import ismethod

from collections import Hashable

from types import FunctionType, MethodType

try:
    from threading import Timer
except ImportError:
    from dummy_threading import Timer

__B3J0F__PROPERTIES__ = '__b3j0f_props'  #: __dict__ properties key

__DICT__ = '__dict__'  #: __dict__ elt attribute name
__CLASS__ = '__class__'  #: __class__ elt attribute name
__SELF__ = '__self__'  #: __self__ class instance attribute name
__BASES__ = '__bases__'  # __bases__ class attribute name

__STATIC_ELEMENTS_CACHE__ = {}  #: dictionary of properties for static objects
__UNHASHABLE_ELTS_CACHE__ = {}  #: dictionary of properties for unhashable obj


def find_ctx(elt):
    """Get the right ctx related to input elt.

    In order to keep safe memory as much as possible, it is important to find
    the right context element. For example, instead of putting properties on
    a function at the level of an instance, it is important to save such
    property on the instance because the function.__dict__ is shared with
    instance class function, and so, if the instance is deleted from memory,
    the property is still present in the class memory. And so on, it is
    impossible to identify the right context in such case if all properties
    are saved with the same key in the same function which is the function.
    """

    result = elt  # by default, result is ctx

    # if elt is ctx and elt is a method, it is possible to find the best ctx
    if ismethod(elt):
        # get instance and class of the elt
        instance = elt.__self__
        # if instance is not None, the right context is the instance
        if instance is not None:
            result = instance

        elif PY2:
            result = elt.im_class

    return result


def free_cache(ctx, *elts):
    """Free properties bound to input cached elts. If empty, free the whole
    cache.
    """

    global __STATIC_ELEMENTS_CACHE__, __UNHASHABLE_ELTS_CACHE__

    for elt in elts:
        if isinstance(elt, Hashable):
            cache = __STATIC_ELEMENTS_CACHE__
        else:
            cache = __UNHASHABLE_ELTS_CACHE__
            elt = id(elt)
        if elt in cache:
            del cache[elt]

    if not elts:
        __STATIC_ELEMENTS_CACHE__ = {}
        __UNHASHABLE_ELTS_CACHE__ = {}


def _ctx_elt_properties(elt, ctx=None, create=False):
    """Get elt properties related to a ctx.

    :param elt: property component elt. Not None methods or unhashable types.
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.
    :param bool create: create ctx elt properties if not exist.

    :return: dictionary of property by name embedded into ctx __dict__ or in
        shared __STATIC_ELEMENTS_CACHE__ or else in shared
        __UNHASHABLE_ELTS_CACHE__. None if no properties exists.
    :rtype: dict
    """

    result = None

    if ctx is None:
        ctx = find_ctx(elt=elt)

    ctx_properties = None

    # in case of dynamic object, parse the ctx __dict__
    ctx__dict__ = getattr(ctx, __DICT__, None)
    if ctx__dict__ is not None and isinstance(ctx__dict__, dict):
        # if properties exist in ctx__dict__
        if __B3J0F__PROPERTIES__ in ctx__dict__:
            ctx_properties = ctx__dict__[__B3J0F__PROPERTIES__]
        elif create:  # if create in worst case
            ctx_properties = ctx__dict__[__B3J0F__PROPERTIES__] = {}

    else:  # in case of static object
        if isinstance(ctx, Hashable):  # search among static elements
            cache = __STATIC_ELEMENTS_CACHE__
        else:  # or unhashable elements
            cache = __UNHASHABLE_ELTS_CACHE__
            ctx = id(ctx)
            if not isinstance(elt, Hashable):
                elt = id(elt)
        # if ctx is in cache
        if ctx in cache:
            # get its properties
            ctx_properties = cache[ctx]
        elif create:  # elif create, get an empty dict
            ctx_properties = cache[ctx] = {}

    # if ctx_properties is not None
    if ctx_properties is not None:
        # check if elt exist in ctx_properties
        if elt in ctx_properties:
            result = ctx_properties[elt]
        elif create:  # create the right data if create
            if create:
                result = ctx_properties[elt] = {}

    return result


def get_properties(elt, keys=None, ctx=None):
    """Get elt properties.

    :param elt: properties elt. Not None methods or unhashable types.
    :param keys: key(s) of properties to get from elt.
        If None, get all properties.
    :type keys: list or str
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.

    :return: list of properties by elt and name.
    :rtype: list
    """

    # initialize keys if str
    if isinstance(keys, str):
        keys = (keys,)

    result = _get_properties(elt, keys=keys, local=False, ctx=ctx)

    return result


def get_property(elt, key, ctx=None):
    """Get elt key property.

    :param elt: property elt. Not None methods.
    :param key: property key to get from elt.
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.

    :return: list of property values by elt.
    :rtype: list
    """

    result = []

    properties = get_properties(elt=elt, ctx=ctx, keys=key)

    if key in properties:
        result = properties[key]

    return result


def get_first_properties(elt, keys=None, ctx=None):
    """Get first properties related to one input key.

    :param elt: first property elt. Not None methods.
    :param list keys: property keys to get.
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.

    :return: dict of first values of elt properties.
    """

    # ensure keys is an iterable if not None
    if isinstance(keys, str):
        keys = (keys,)

    result = _get_properties(elt, keys=keys, first=True, ctx=ctx)

    return result


def get_first_property(elt, key, default=None, ctx=None):
    """Get first property related to one input key.

    :param elt: first property elt. Not None methods.
    :param str key: property key to get.
    :param default: default value to return if key does not exist in elt.
        properties
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.
    """

    result = default

    properties = _get_properties(elt, keys=(key,), ctx=ctx, first=True)

    # set value if key exists in properties
    if key in properties:
        result = properties[key]

    return result


def get_local_properties(elt, keys=None, ctx=None):
    """Get local elt properties (not defined in elt type or base classes).

    :param elt: local properties elt. Not None methods.
    :param keys: keys of properties to get from elt.
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.

    :return: dict of properties by name.
    :rtype: dict
    """

    if isinstance(keys, str):
        keys = (keys,)

    result = _get_properties(elt, keys=keys, local=True, ctx=ctx)

    return result


def get_local_property(elt, key, default=None, ctx=None):
    """Get one local property related to one input key or default value if key
    is not found.

    :param elt: local property elt. Not None methods.
    :param str key: property key to get.
    :param default: default value to return if key does not exist in elt
        properties.
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.
    :return: dict of properties by name.
    :rtype: dict
    """

    result = default

    local_properties = get_local_properties(elt=elt, keys=(key,), ctx=ctx)

    if key in local_properties:
        result = local_properties[key]

    return result


def _get_properties(
    elt, keys=None, local=False, exclude=None, first=False, ctx=None,
    _found_keys=None
):
    """Get a dictionary of elt properties.

    Such dictionary is filled related to the elt properties, and from all base
        elts properties if not local.

    :param elt: properties elt. Not None methods.
    :param ctx: elt ctx from where get properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.
    :param keys: keys of properties to get from elt. If None, get all
        properties.
    :param bool local: if False, get properties from bases classes and type
        as well.
    :param set exclude: elts from where not get properties.
    :param bool first: get only first property values.
    :param _found_keys: used such as a cache value for result.

    :return: dict of properties by name and ...:

        - if local or first: value.
        - other cases: [(elt, value)] in order of property definition even if
            multi inheritance do not ensure to respect inheritance order in a
            list.
    :rtype: dict
    """

    result = {}

    # get the best context
    if ctx is None:
        ctx = find_ctx(elt=elt)

    # initialize found keys
    if first and _found_keys is None:
        _found_keys = set()

    # get elt properties if exist
    elt_properties = _ctx_elt_properties(elt=elt, ctx=ctx, create=False)

    # if elt_properties exists
    if elt_properties is not None:
        # if elt properties is not empty
        # try to add property values in result related to input keys
        properties_to_keep = elt_properties.copy() if keys is None else {}
        # get key values
        if keys is not None:
            for key in keys:
                if key in elt_properties:
                    if not first or key not in _found_keys:
                        properties_to_keep[key] = elt_properties[key]
                    if first and key not in _found_keys:
                        _found_keys.add(key)

        if local:  # in case of local, use properties_to_keep
            result = properties_to_keep

        # if not local and not first and properties to keep
        elif not first and properties_to_keep:
            for name in properties_to_keep:
                property_to_keep = properties_to_keep[name]
                # add (elt, properties_to_keep) in result
                if name not in result:
                    result[name] = [(elt, property_to_keep)]
                else:
                    result[name].append((elt, property_to_keep))

        # in case of first, result is properties_to_keep
        elif properties_to_keep:
            result = properties_to_keep
    # check if all keys have been found if first
    keys_left = not first or keys is None or len(keys) != len(_found_keys)

    # if not local or keys left
    if not local and keys_left:
        # initialize exclude
        if exclude is None:
            exclude = set()

        # add elt in exclude in order to avoid to get elt properties twice
        if isinstance(elt, Hashable):
            exclude.add(elt)

        # and search among bases ctx elements
        if ctx is not elt:
            # go to base elts
            try:
                elt_name = elt.__name__
            except AttributeError:
                pass
            else:
                # if elt is a sub attr of ctx
                if hasattr(ctx, elt_name):
                    if hasattr(ctx, __BASES__):
                        ctx_bases = ctx.__bases__
                    elif hasattr(ctx, __CLASS__):
                        ctx_bases = (ctx.__class__,)
                    else:
                        ctx_bases = ()
                    for base_ctx in ctx_bases:
                        # get base_elt properties
                        base_elt = getattr(base_ctx, elt_name, None)
                        if base_elt is not None:
                            base_properties = _get_properties(
                                elt=base_elt,
                                keys=keys,
                                local=local,
                                exclude=exclude,
                                ctx=base_ctx,
                                _found_keys=_found_keys
                            )
                            # update result with base_properties
                            for name in base_properties:
                                properties = base_properties[name]
                                if name not in result:
                                    result[name] = properties
                                else:
                                    result[name] += properties

        else:
            # bases classes
            if hasattr(elt, __BASES__):
                for base in elt.__bases__:
                    if base not in exclude:
                        base_result = _get_properties(
                            elt=base,
                            keys=keys,
                            local=local,
                            exclude=exclude,
                            ctx=base,
                            _found_keys=_found_keys
                        )
                        # update result with base_result
                        for name in base_result:
                            properties = base_result[name]
                            if name not in result:
                                result[name] = properties
                            else:
                                result[name] += properties

            # search among type definition

            # class
            if hasattr(elt, __CLASS__):
                elt_class = elt.__class__
                # get class properties only if elt is not its own class
                if elt_class is not elt and elt_class not in exclude:
                    elt_class_properties = _get_properties(
                        elt=elt_class,
                        keys=keys,
                        local=local,
                        exclude=exclude,
                        ctx=elt_class,
                        _found_keys=_found_keys
                    )
                    # update result with class result
                    for name in elt_class_properties:
                        properties = elt_class_properties[name]
                        if name not in result:
                            result[name] = properties
                        else:
                            result[name] += properties

            # type
            elt_type = type(elt)
            # get elt type properties only if elt is not its type
            if elt_type is not elt and elt_type not in exclude:
                elt_type_properties = _get_properties(
                    elt=elt_type,
                    keys=keys,
                    local=local,
                    exclude=exclude,
                    ctx=elt_type,
                    _found_keys=_found_keys
                )
                # update result with type properties
                for name in elt_type_properties:
                    properties = elt_type_properties[name]
                    if name not in result:
                        result[name] = properties
                    else:
                        result[name] += properties

    return result


def put_property(elt, key, value, ttl=None, ctx=None):
    """Put properties in elt.

    :param elt: properties elt to put. Not None methods.
    :param number ttl: If not None, property time to leave.
    :param ctx: elt ctx from where put properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.
    :param dict properties: properties to put in elt. elt and ttl are exclude.

    :return: Timer if ttl is not None.
    :rtype: Timer
    """
    return put_properties(elt=elt, properties={key: value}, ttl=ttl, ctx=ctx)


def put_properties(elt, properties, ttl=None, ctx=None):
    """Put properties in elt.

    :param elt: properties elt to put. Not None methods.
    :param number ttl: If not None, property time to leave.
    :param ctx: elt ctx from where put properties. Equals elt if None. It
        allows to get function properties related to a class or instance if
        related function is defined in base class.
    :param dict properties: properties to put in elt. elt and ttl are exclude.

    :return: Timer if ttl is not None.
    :rtype: Timer
    """

    result = None

    if properties:

        # get the best context
        if ctx is None:
            ctx = find_ctx(elt=elt)

        # get elt properties
        elt_properties = _ctx_elt_properties(elt=elt, ctx=ctx, create=True)

        # set properties
        elt_properties.update(properties)

        if ttl is not None:
            kwargs = {
                'elt': elt,
                'ctx': ctx,
                'keys': tuple(properties.keys())
            }
            result = Timer(ttl, del_properties, kwargs=kwargs)
            result.start()

    return result


def put(properties, ttl=None, ctx=None):
    """Decorator dedicated to put properties on an element.
    """

    def put_on(elt):
        return put_properties(elt=elt, properties=properties, ttl=ttl, ctx=ctx)

    return put_on


def del_properties(elt, keys=None, ctx=None):
    """Delete elt property.

    :param elt: properties elt to del. Not None methods.
    :param keys: property keys to delete from elt. If empty, delete all
        properties.
    """

    # get the best context
    if ctx is None:
        ctx = find_ctx(elt=elt)

    elt_properties = _ctx_elt_properties(elt=elt, ctx=ctx, create=False)

    # if elt properties exist
    if elt_properties is not None:

        if keys is None:
            keys = list(elt_properties.keys())

        else:
            keys = ensureiterable(keys, iterable=tuple, exclude=str)

        for key in keys:
            if key in elt_properties:
                del elt_properties[key]

        # delete property component if empty
        if not elt_properties:
            # case of dynamic object
            if isinstance(getattr(ctx, '__dict__', None), dict):
                try:
                    if elt in ctx.__dict__[__B3J0F__PROPERTIES__]:
                        del ctx.__dict__[__B3J0F__PROPERTIES__][elt]
                except TypeError:  # if elt is unhashable
                    elt = id(elt)
                    if elt in ctx.__dict__[__B3J0F__PROPERTIES__]:
                        del ctx.__dict__[__B3J0F__PROPERTIES__][elt]
                # if ctx_properties is empty, delete it
                if not ctx.__dict__[__B3J0F__PROPERTIES__]:
                    del ctx.__dict__[__B3J0F__PROPERTIES__]

            # case of static object and hashable
            else:
                if isinstance(ctx, Hashable):
                    cache = __STATIC_ELEMENTS_CACHE__
                else:  # case of static and unhashable object
                    cache = __UNHASHABLE_ELTS_CACHE__
                    ctx = id(ctx)
                    if not isinstance(elt, Hashable):
                        elt = id(elt)

                # in case of static object
                if ctx in cache:
                    del cache[ctx][elt]
                    if not cache[ctx]:
                        del cache[ctx]


def firsts(properties):
    """
    Transform a dictionary of {name: [(elt, value)+]} (resulting from
        get_properties) to a dictionary of {name, value} where names are first
        encountered in input properties.

    :param dict properties: properties to firsts.
    :return: dictionary of parameter values by names.
    :rtype: dict
    """

    result = {}

    # parse elts
    for name in properties:
        elt_properties = properties[name]
        # add property values in result[name]
        result[name] = elt_properties[0][1]

    return result


def remove_ctx(properties):
    """
    Transform a dictionary of {name: [(elt, value)+]} into a dictionary of
        {name: [value+]}.

    :param dict properties: properties from where get only values.
    :return: dictionary of parameter values by names.
    :rtype: dict
    """
    result = {}

    for name in properties:
        elt_properties = properties[name]
        result[name] = []
        for _, value in elt_properties:
            result[name].append(value)

    return result


def setdefault(elt, key, default, ctx=None):
    """
    Get a local property and create default value if local property does not
        exist.

    :param elt: local proprety elt to get/create. Not None methods.
    :param str key: proprety name.
    :param default: property value to set if key no in local properties.

    :return: property value or default if property does not exist.
    """

    result = default

    # get the best context
    if ctx is None:
        ctx = find_ctx(elt=elt)

    # get elt properties
    elt_properties = _ctx_elt_properties(elt=elt, ctx=ctx, create=True)

    # if key exists in elt properties
    if key in elt_properties:
        # result is elt_properties[key]
        result = elt_properties[key]
    else:  # set default property value
        elt_properties[key] = default

    return result


def _protectedattrname(name):
    """Get protected attribute name from input public attribute name.

    :param str name: public attribute name.
    :return: protected attribute name.
    :rtype: str
    """

    return '_{0}'.format(name)


def addproperties(
        names,
        bfget=None, afget=None, enableget=True,
        bfset=None, afset=None, enableset=True,
        bfdel=None, afdel=None, enabledel=True
):
    """Decorator in charge of adding python properties to cls.

    {a/b}fget, {a/b}fset and {a/b}fdel are applied to all properties matching
    names in taking care to not forget default/existing properties. The
    prefixes *a* and *b* are respectively for after and before default/existing
    property getter/setter/deleter execution.

    These getter, setter and deleter functions are called before existing or
    default getter, setter and deleters. Default getter, setter and deleters
    are functions which uses an attribute with a name starting with '_' and
    finishing with the property name (like the python language convention).

    .. seealso::
        _protectedattrname(name)

    :param str(s) names: property name(s) to add.
    :param bfget: getter function to apply to all properties before
        default/existing getter execution. Parameters are a decorated cls
        instance and a property name.
    :param afget: fget function to apply to all properties after
        default/existing getter execution. Parameters are a decorated cls
        instance and a property name.
    :param bool enableget: if True (default), enable existing or default getter
        . Otherwise, use only fget if given.
    :param bfset: fset function to apply to all properties before
        default/existing setter execution. Parameters are a decorated cls
        instance and a property name.
    :param afset: fset function to apply to all properties after
        default/existing setter execution. Parameters are a decorated cls
        instance and a property name.
    :param bool enableset: if True (default), enable existing or default setter
        . Otherwise, use only fset if given.
    :param bfdel: fdel function to apply to all properties before
        default/existing deleter execution. Parameters are a decorated cls
        instance and a property name.
    :param bfdel: fdel function to apply to all properties after
        default/existing deleter execution. Parameters are a decorated cls
        instance and a property name.
    :param bool enabledel: if True (default), enable existing or default
        deleter. Otherwise, use only fdel if given.
    :return: cls decorator.
    """

    names = ensureiterable(names, exclude=str)  # ensure names is a list

    if isinstance(bfget, MethodType):
        finalbfget = lambda self, name: bfget(name)
    else:
        finalbfget = bfget

    if isinstance(afget, MethodType):
        finalafget = lambda self, name: afget(name)
    else:
        finalafget = afget

    if isinstance(bfset, MethodType):
        finalbfset = lambda self, value, name: bfset(value, name)
    else:
        finalbfset = bfset

    if isinstance(afset, MethodType):
        finalafset = lambda self, value, name: afset(value, name)
    else:
        finalafset = afset

    if isinstance(bfdel, MethodType):
        finalbfdel = lambda self, name: bfdel(name)
    else:
        finalbfdel = bfdel

    if isinstance(afdel, MethodType):
        finalafdel = lambda self, name: afdel(name)
    else:
        finalafdel = afdel

    def addproperties(cls):
        """Add properties to cls.

        :param type cls: cls on adding properties.
        :return: cls
        """

        for name in names:
            # try to find an existing property
            existingproperty = getattr(cls, name, None)
            if isinstance(existingproperty, property):
                _fget = existingproperty.fget
                _fset = existingproperty.fset
                _fdel = existingproperty.fdel

            else:
                _fget, _fset, _fdel = None, None, None

            # construct existing/default getter
            if _fget is None:
                def _fget(name):
                    """Simple getter wrapper."""
                    def _fget(self):
                        """Simple getter."""
                        return getattr(self, _protectedattrname(name))
                    return _fget
                _fget = _fget(name)
            # transform method to function in order to add self in parameters
            if isinstance(_fget, MethodType):
                final_fget = lambda self: _fget()
            else:
                final_fget = _fget

            # construct existing/default setter
            if _fset is None:
                def _fset(name):
                    """Simple setter wrapper."""
                    def _fset(self, value):
                        """Simple setter."""
                        setattr(self, _protectedattrname(name), value)
                    return _fset
                _fset = _fset(name)
            # transform method to function in order to add self in parameters
            if isinstance(_fset, MethodType):
                final_fset = lambda self, value: _fset(value)
            else:
                final_fset = _fset

            # construct existing/default deleter
            if _fdel is None:
                def _fdel(name):
                    """Simple deleter wrapper."""
                    def _fdel(self):
                        """Simple deleter."""
                        delattr(self, _protectedattrname(name))
                    return _fdel
                _fdel = _fdel(name)
            # transform method to function in order to add self in parameters
            if isinstance(_fdel, MethodType):
                final_fdel = lambda self: _fdel()
            else:
                final_fdel = _fdel

            def _getter(final_fget, name):
                """Property getter wrapper."""
                def _getter(self):
                    """Property getter."""
                    result = None
                    # start to process input bfget
                    if finalbfget is not None:
                        result = finalbfget(self, name)
                    # process cls getter
                    if enableget:
                        result = final_fget(self)
                    # finish to process afget
                    if finalafget is not None:
                        result = finalafget(self, name)
                    return result
                return _getter
            _getter = _getter(final_fget, name)

            def _setter(final_fset, name):
                """Property setter wrapper."""
                def _setter(self, value):
                    """Property setter."""
                    # start to process input bfset
                    if finalbfset is not None:
                        finalbfset(self, value, name)
                    # finish to process cls setter
                    if enableset:
                        final_fset(self, value)
                    # finish to process afset
                    if finalafset is not None:
                        finalafset(self, value, name)
                return _setter
            _setter = _setter(final_fset, name)

            def _deleter(final_fdel, name):
                """Property deleter wrapper."""
                def _deleter(self):
                    """Property deleter."""
                    # start to process input fdel
                    if finalbfdel is not None:
                        finalbfdel(self, name)
                    # finish to process cls deleter
                    if enabledel:
                        final_fdel(self)
                    # finish to process afget
                    if finalafdel is not None:
                        finalafdel(self, name)
                return _deleter
            _deleter = _deleter(final_fdel, name)

            # get property name
            propertyfield = property(fget=_getter, fset=_setter, fdel=_deleter)

            # put property name in cls
            setattr(cls, name, propertyfield)

        return cls  # finish to return the cls

    return addproperties



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase
from b3j0f.utils.path import lookup, clearcache, incache, getpath


class LookUpTest(UTCase):

    def setUp(self):
        pass

    def test_builtin(self):
        """
        Test lookup of builtin
        """

        _open = lookup('%s.open' % open.__module__)
        self.assertIs(_open, open)

    def test_package(self):
        """
        Test lookup package
        """

        package_path = 'b3j0f'
        package = lookup(package_path)
        self.assertEqual(package.__name__, package_path)

    def test_module(self):
        """
        Test lookup of module
        """

        module_path = 'b3j0f.utils'
        module = lookup(module_path)
        self.assertEqual(module.__name__, module_path)

    def test_class(self):
        """
        Test lookup class
        """

        cls = lookup('b3j0f.utils.test.path.LookUpTest')
        self.assertIs(cls, LookUpTest)

    def test_function(self):
        """
        Test lookup function
        """

        f = lookup('b3j0f.utils.path.lookup')
        self.assertIs(f, lookup)

    def test_method(self):
        """
        Test lookup method
        """

        m = lookup('b3j0f.utils.test.path.LookUpTest.test_method')
        self.assertEqual(m, LookUpTest.test_method)

    def test_local(self):
        """
        Test lookup local
        """

        def f_test():
            pass

        l = lookup('f_test')
        self.assertIs(l, f_test)

    def test_notfound(self):
        """
        Test lookup on not existing element
        """

        self.assertRaises(ImportError, lookup, 'unexist')

    def test_cache(self):
        """
        Test lookup cache
        """

        # check if cache is empty
        path = 'b3j0f'
        self.assertFalse(incache(path))
        # check if element is saved in the cache
        lookup(path, cache=True)
        self.assertTrue(incache(path))
        # check if clear cache works
        clearcache(path)
        self.assertFalse(incache(path))
        # check if clear all cache works
        lookup(path, cache=True)
        clearcache()
        self.assertFalse(incache(path))
        # check if element is not saved in the cache
        lookup(path, cache=False)
        self.assertFalse(incache(path))


class GetPathTest(UTCase):

    def test_builtin(self):
        """
        Test getpath builtin
        """

        open_path = getpath(open)
        self.assertEqual(open_path, '%s.open' % open.__module__)

    def test_class(self):
        """
        Test getpath class
        """

        cls_path = getpath(GetPathTest)
        self.assertEqual(cls_path, 'b3j0f.utils.test.path.GetPathTest')


class LookUpGetPathTest(UTCase):

    def test_lookup_getpath(self):
        """
        Test lookup + getpath
        """

        path = '%s.open' % open.__module__
        open_path = getpath(lookup(path))
        self.assertEqual(open_path, path)

    def test_getpath_lookup(self):
        """
        Test getpath + lookup
        """

        f = open
        _open = lookup(getpath(f))
        self.assertIs(_open, f)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Provides tools to manage iterable types
"""

__all__ = ['isiterable', 'ensureiterable', 'first']

from collections import Iterable


def isiterable(element, exclude=None):
    """Check whatever or not if input element is an iterable.

    :param element: element to check among iterable types.
    :param type/tuple exclude: not allowed types in the test.

    :Example:

    >>> isiterable({})
    True
    >>> isiterable({}, exclude=dict)
    False
    >>> isiterable({}, exclude=(dict,))
    False
    """

    # check for allowed type
    allowed = exclude is None or not isinstance(element, exclude)
    result = allowed and isinstance(element, Iterable)

    return result


def ensureiterable(value, iterable=list, exclude=None):
    """Convert a value into an iterable if it is not.

    :param object value: object to convert
    :param type iterable: iterable type to apply (default: list)
    :param type/tuple exclude: types to not convert

    :Example:

    >>> ensureiterable([])
    []
    >>> ensureiterable([], iterable=tuple)
    ()
    >>> ensureiterable('test', exclude=str)
    ['test']
    >>> ensureiterable('test')
    ['t', 'e', 's', 't']
    """

    result = value

    if not isiterable(value, exclude=exclude):
        result = [value]
        result = iterable(result)

    else:
        result = iterable(value)

    return result


def first(iterable, default=None):
    """Try to get input iterable first item or default if iterable is empty.

    :param Iterable iterable: iterable to iterate on.
    :param default: default value to get if input iterable is empty.
    :raises TypeError: if iterable is not an iterable value

    :Example:

    >>> first('test')
    't'
    >>> first([])
    None
    >>> first('', default='test')
    'test'
    """

    # start to get the iterable iterator (raises TypeError if iter)
    iterator = iter(iterable)
    # get first element
    try:
        result = next(iterator)
    except StopIteration:
        # if no element exist, result equals default
        result = default

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase
from b3j0f.utils.iterable import first, ensureiterable, isiterable
from b3j0f.utils.version import basestring


class EnsureIterableTest(UTCase):

    def test_list(self):

        l = []
        itererable = ensureiterable(l)
        self.assertEqual(itererable, l)

    def test_dict(self):

        l = []
        iterable = ensureiterable(l, iterable=dict)
        self.assertTrue(isinstance(iterable, dict))
        self.assertFalse(iterable)

    def test_exclude(self):

        l = ""
        iterable = ensureiterable(l, exclude=basestring)
        self.assertTrue(iterable)


class IsIterable(UTCase):

    def test_iterable(self):
        """
        Test an iterable value
        """

        self.assertTrue(isiterable([]))

    def test_exclude(self):
        """
        Test iterable and not allowed types
        """

        self.assertFalse(isiterable([], exclude=list))

    def test_excludes(self):
        """
        Test iterable with a tuple of exclude types
        """

        self.assertFalse(isiterable([], exclude=(list, basestring)))

    def test_not_iterable(self):
        """
        Test not iterable element
        """

        self.assertFalse(isiterable(None))


class First(UTCase):

    def test_empty(self):
        """
        Test empty iterable
        """

        self.assertIs(first([]), None)

    def test_first(self):
        """
        Test first element in an iterable
        """

        value = 'value'
        iterable = [value]

        self.assertIs(first(iterable), value)

    def test_default(self):
        """
        Test default value
        """

        default = 'default'

        self.assertIs(first([], default), default)

    def test_notiterable(self):
        """
        Test not iterable
        """

        self.assertRaises(TypeError, first, True)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase
from b3j0f.utils.reflect import base_elts, find_embedding

from inspect import getmodule


class BaseEltsTest(UTCase):
    """
    Test base_elts function
    """

    def test_not_inherited(self):
        """
        Test with a not inherited element.
        """

        bases = base_elts(None)
        self.assertFalse(bases)

    def test_function(self):
        """
        Test function
        """

        bases = base_elts(lambda: None)
        self.assertFalse(bases)

    def test_class(self):
        """
        Test class
        """

        class A:
            pass

        class B(A, dict):
            pass

        bases = base_elts(B)
        self.assertEqual(bases, list(B.__bases__) + [object])

    def test_method(self):
        """
        Test method
        """

        class A:
            def a(self):
                pass

        class B(A):
            pass

        bases = base_elts(B.a, cls=A)
        self.assertEqual(len(bases), 1)
        base = bases.pop()
        self.assertEqual(base, A.a)

    def test_not_method(self):
        """
        Test when method has been overriden
        """

        class A:
            def a(self):
                pass

        class B(A):
            def a(self):
                pass

        bases = base_elts(B.a, cls=A)
        self.assertFalse(bases)

    def test_boundmethod(self):
        """
        Test bound method
        """

        class Test:
            def test(self):
                pass

        test = Test()

        bases = base_elts(test.test)
        self.assertEqual(len(bases), 1)
        self.assertEqual(bases.pop(), Test.test)

    def test_not_boundmethod(self):
        """
        Test with a bound method which is only defined in the instance
        """

        class Test:
            def test(self):
                pass

        test = Test()
        test.test = lambda self: None

        bases = base_elts(test.test)
        self.assertFalse(bases)


class FindEmbeddingTest(UTCase):

    def test_none(self):

        embedding = find_embedding(None)

        self.assertFalse(embedding)

    def test_wrong_embedding(self):

        embedding = find_embedding(None, embedding=FindEmbeddingTest)

        self.assertFalse(embedding)

    def test_module(self):

        FindEmbeddingTestModule = getmodule(FindEmbeddingTest)
        embedding = find_embedding(FindEmbeddingTestModule)

        self.assertEqual(len(embedding), 1)
        self.assertIs(FindEmbeddingTestModule, embedding[0])

    def test_function(self):

        embedding = find_embedding(find_embedding)

        self.assertEqual(len(embedding), 2)
        self.assertIs(embedding[0], getmodule(find_embedding))
        self.assertIs(embedding[1], find_embedding)

    def test_class(self):

        embedding = find_embedding(FindEmbeddingTest)

        self.assertEqual(len(embedding), 2)
        self.assertIs(getmodule(FindEmbeddingTest), embedding[0])
        self.assertIs(FindEmbeddingTest, embedding[1])

    def test_method(self):

        embedding = find_embedding(FindEmbeddingTest.test_method)

        self.assertEqual(len(embedding), 3)
        self.assertIs(getmodule(FindEmbeddingTest), embedding[0])
        self.assertIs(FindEmbeddingTest, embedding[1])
        self.assertEqual(FindEmbeddingTest.test_method, embedding[2])

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from time import sleep

from random import random

from b3j0f.utils.ut import UTCase
from b3j0f.utils.version import PY2
from b3j0f.utils.property import (
    get_properties, put_properties, del_properties,
    get_local_properties, get_local_property,
    firsts, remove_ctx,
    get_first_property, get_first_properties,
    setdefault, put_property,
    __B3J0F__PROPERTIES__,
    find_ctx, addproperties, _protectedattrname
)


class FindCTXTest(UTCase):
    """
    Test find_ctx method
    """
    def test_notctx(self):
        """
        Test with elt without ctx
        """

        elt = 1

        ctx = find_ctx(elt=elt)

        self.assertIs(ctx, elt)

    def test_method(self):
        """
        Test method ctx
        """

        class A:
            def a(self):
                pass

        elt = A.a

        ctx = find_ctx(elt=elt)

        if PY2:
            self.assertIs(A, ctx)
        else:
            self.assertIs(elt, ctx)

    def test_instance_method(self):
        """
        Test instance method ctx
        """

        class A:
            def a(self):
                pass

        a = A()

        elt = a.a

        ctx = find_ctx(elt=elt)

        self.assertIs(a, ctx)


class RemoveCTXTest(UTCase):
    """
    Test remove_ctx function.
    """

    def test_no_properties(self):
        """
        Test if not properties exist.
        """

        properties = remove_ctx({})

        self.assertFalse(properties)

    def test_empty(self):
        """
        Test with empty properties.
        """

        properties = remove_ctx({'test': []})

        self.assertFalse(properties['test'])

    def test_one(self):
        """
        Test with one value property.
        """

        properties = {'test': [('elt', 0)]}

        properties = remove_ctx(properties)

        self.assertEqual(properties['test'], [0])

    def test_properties(self):
        """
        Test with many property values.
        """

        properties = {'test': [('elt', 0), ('elt', 1)]}

        properties = remove_ctx(properties)

        self.assertEqual(properties['test'], [0, 1])


class PropertyTest(UTCase):
    """
    Test scenarios of puting/getting/deleting properties.
    """

    def _assert_properties(self, elt, count=10):

        properties = get_properties(elt=elt)
        self.assertFalse(properties)
        local_properties = get_local_properties(elt=elt)
        self.assertFalse(local_properties)

        properties = dict((str(i), i) for i in range(count))
        put_properties(elt=elt, properties=properties)

        local_properties = get_local_properties(elt=elt)
        properties = get_properties(elt=elt)

        self.assertEqual(len(properties), count)
        self.assertEqual(len(local_properties), count)
        for index in range(count):
            name = str(index)
            self.assertIs(properties[name][0][0], elt)
            self.assertIs(properties[name][0][1], index)
            self.assertIs(local_properties[name], index)

        for index in range(count):
            name = str(index)
            properties = get_properties(elt=elt, keys=name)
            self.assertEqual(len(properties), 1)
            self.assertIs(properties[name][0][0], elt)
            self.assertIn(name, properties)
            self.assertIs(properties[name][0][1], index)
            local_properties = get_local_properties(elt=elt, keys=name)
            self.assertEqual(len(local_properties), 1)
            self.assertIn(name, local_properties)
            self.assertIs(local_properties[name], index)

        del_properties(elt=elt, keys='0')

        properties = get_properties(elt=elt)
        self.assertEqual(len(properties), count - 1)

        local_properties = get_local_properties(elt=elt)
        self.assertEqual(len(local_properties), count - 1)

        for i in range(1, count):
            name = str(i)
            self.assertIs(properties[name][0][1], i)
            self.assertIs(local_properties[name], i)

        del_properties(elt=elt)
        properties = get_properties(elt=elt)
        self.assertFalse(properties)
        local_properties = get_local_properties(elt=elt)
        self.assertFalse(local_properties)

        if hasattr(elt, '__dict__'):
            self.assertNotIn(__B3J0F__PROPERTIES__, elt.__dict__)

    def test_builtin(self):
        """
        Test lookup of builtin
        """

        self._assert_properties(min)

    def test_object(self):
        """
        Test scenario on an object
        """
        self._assert_properties(1)

    def test_none(self):
        """
        Test scenario on None
        """
        self._assert_properties(None)

    def test_lambda(self):
        """
        Test scenario on a lambda expression
        """
        self._assert_properties(lambda: None)

    def test_function(self):
        """
        Test scenario on a function.
        """
        def a():
            pass

        self._assert_properties(a)

    def _test_inheritance(self, first, second, ctx1=None, ctx2=None, count=5):
        """
        Test inherited properties between first and second elements.
        """

        properties = dict((str(i), i) for i in range(count))
        put_properties(elt=first, properties=properties, ctx=ctx1)

        properties = get_properties(elt=second, ctx=ctx2)
        self.assertEqual(len(properties), count)
        self.assertEqual(first, properties['0'][0][0])

        local_properties = get_local_properties(elt=second, ctx=ctx2)
        self.assertEqual(len(local_properties), 0)

        properties = dict((str(i), i) for i in range(count))
        put_properties(elt=second, properties=properties, ctx=ctx2)

        properties = get_properties(elt=second, ctx=ctx2)
        self.assertEqual(len(properties), count)
        self.assertEqual(first, properties['0'][1][0])
        self.assertEqual(second, properties['0'][0][0])

        local_properties = get_local_properties(elt=second, ctx=ctx2)
        self.assertEqual(len(local_properties), count)

        del_properties(elt=first, ctx=ctx1)
        properties = get_properties(elt=second, ctx=ctx2)
        self.assertEqual(len(properties), count)
        self.assertEqual(len(properties['0']), 1)
        self.assertIs(second, properties['0'][0][0])
        local_properties = get_local_properties(elt=second, ctx=ctx2)
        self.assertEqual(len(local_properties), count)

        del_properties(elt=second, ctx=ctx2)
        properties = get_properties(elt=second, ctx=ctx2)
        self.assertFalse(properties)
        local_properties = get_local_properties(elt=second, ctx=ctx2)
        self.assertFalse(local_properties)

    def test_class(self):
        """
        Test scenario on a class.
        """
        class A(object):
            pass

        self._assert_properties(A)

        class B(A):
            pass

        self._test_inheritance(A, B)

    def test_instance(self):
        """
        Test scenario on an instance.
        """
        class A:
            pass

        a = A()

        self._assert_properties(a)

        self._test_inheritance(A, a)

    def test_namespace(self):
        """
        Test scenario on a namespace.
        """
        class A:
            pass

        class B(A):
            pass

        self._assert_properties(A)

        self._test_inheritance(A, B)

    def test_method(self):
        """
        Test scenario on a method.
        """
        class A:
            def a(self):
                pass

        class B(A):
            pass

        self._assert_properties(A.a)

        self._test_inheritance(A.a, B.a, ctx1=A, ctx2=B)

    def test_bound_method(self):
        """
        Test scenario on a bound method.
        """
        class A(object):
            def a(self):
                pass

        a = A()

        self._test_inheritance(A.a, a.a, ctx1=A)

    def test_module(self):
        """
        Test scenario on a module.
        """
        import b3j0f

        self._assert_properties(b3j0f)

    def test_property_module(self):
        """
        Test scenario on the property module.
        """
        import b3j0f.utils.property

        self._assert_properties(b3j0f.utils.property)

    def test_dict(self):
        """
        Test scenario on a dictionary.
        """
        elt = {}

        self._assert_properties(elt)

    def test_list(self):
        """
        Test scenario on a list.
        """
        elt = []

        self._assert_properties(elt)

    def test_inheritance(self):
        """
        Test scenario on inherited methods
        """

        key, a, b, c = 'test', 1, 2, 3

        class A:
            def test(self):
                pass

        class B(A):
            pass

        class C(B):
            def test(self):
                pass

        put_property(elt=A.test, ctx=A, key=key, value=a)
        put_property(elt=B.test, ctx=B, key=key, value=b)
        put_property(elt=C.test, ctx=C, key=key, value=c)

        properties = get_properties(elt=C.test, ctx=C)
        self.assertEqual(len(properties), 1)

        self.assertEqual(properties[key][2], (A.test, a))
        self.assertEqual(properties[key][1], (B.test, b))
        self.assertEqual(properties[key][0], (C.test, c))


class TTLTest(UTCase):
    """
    Test ttl parameters while puting property.
    """

    def tearDown(self):
        """
        Del properties at the end of tests.
        """
        del_properties(elt=self)
        self.assertNotIn(__B3J0F__PROPERTIES__, self.__dict__)

    def test_zero(self):
        """
        Test with ttl = 0
        """

        put_property(elt=self, ttl=0, key='name', value=1)

        sleep(0.1)

        properties = get_local_properties(elt=self)

        self.assertFalse(properties)

    def test_100(self):
        """
        Test with ttl = 100
        """

        ttl = 0.1

        put_property(elt=self, ttl=ttl, key='name', value=1)

        properties = get_local_properties(elt=self)

        self.assertTrue(properties)

        sleep(ttl + 0.2)

        properties = get_local_properties(elt=self)

        self.assertFalse(properties)


class OneTest(UTCase):
    """
    UT for get_local_property, get_first_property and get_first_properties.
    """

    def tearDown(self):
        """
        delete properties.
        """
        del_properties(elt=self)
        self.assertNotIn(__B3J0F__PROPERTIES__, self.__dict__)

    def test_first_none(self):
        """
        test first default.
        """
        _property = get_first_property(elt=self, key='a', default=2)

        self.assertEqual(_property, 2)

    def test_first(self):
        """
        test first on existing property
        """
        put_property(elt=self, key='a', value=1)

        _property = get_first_property(elt=self, key='a', default=2)

        self.assertEqual(_property, 1)

    def test_firsts_none(self):
        """
        test firsts default.
        """
        _properties = get_first_properties(elt=self, keys=['a', 'b'])

        self.assertFalse(_properties)

    def test_firsts(self):
        """
        test firsts on existing property
        """
        properties = {'a': 1, 'b': 2}
        put_properties(elt=self, properties=properties)

        _properties = get_first_properties(elt=self, keys='a')

        self.assertIn('a', _properties)
        self.assertEqual(_properties['a'], 1)

    def test_none_local(self):
        """
        test local default.
        """
        local_property = get_local_property(elt=self, key='a', default=2)

        self.assertEqual(local_property, 2)

    def test_local(self):
        """
        test local with existing property.
        """
        put_property(elt=self, key='a', value=1)

        local_property = get_local_property(elt=self, key='a', default=2)

        self.assertEqual(local_property, 1)


class UnifyTest(UTCase):
    """
    Test firsts function.
    """

    def test_empty(self):
        """
        Test empty properties.
        """
        properties = []

        unified_properties = firsts(properties=properties)

        self.assertFalse(unified_properties)

    def test_one_value(self):
        """
        Test one property
        """

        properties = {1: [(2, 3)]}

        unified_properties = firsts(properties=properties)

        self.assertIn(1, unified_properties)
        self.assertEqual(unified_properties[1], 3)

    def test_values(self):
        """
        Test several properties.
        """

        count = 10

        properties = {}

        for i in range(count):
            properties[i] = []
            for j in range(i + 1):
                properties[i].append((i, j))

        unified_properties = firsts(properties=properties)

        self.assertEqual(len(unified_properties), count)
        for i in range(count):
            self.assertIs(unified_properties[i], 0)


class SetDefaultTest(UTCase):
    """
    Test setdefault function
    """

    def setUp(self):
        """
        Set attributes to self key=test and new_value=2
        """
        self.key = 'test'

        self.new_value = 2

    def tearDown(self):
        """
        del properties
        """
        del_properties(elt=self)

        self.assertNotIn(__B3J0F__PROPERTIES__, self.__dict__)

    def test_exists(self):
        """
        Test with an existing property
        """

        put_properties(elt=self, properties={self.key: self.new_value + 1})

        value = setdefault(elt=self, key=self.key, default=self.new_value)

        self.assertNotEqual(value, self.new_value)

    def test_new(self):
        """
        Test on a missing property
        """

        value = setdefault(elt=self, key=self.key, default=self.new_value)

        self.assertEqual(value, self.new_value)


class TestAddProperties(UTCase):
    """Test the addproperties decorator."""

    def setUp(self):

        self.count = 5
        # generate count random property names
        self.names = [str(random()) for i in range(self.count)]

        self.getternames = set()
        self.setternames = {}
        self.deleternames = set()

    def getter(self, name=None):
        """Property getter."""
        if name is None:
            name = self.name
        self.getternames.add(name)
        return self

    def setter(self, value, name=None):
        """Property setter."""
        if name is None:
            name = self.name
        self.setternames[name] = value

    def deleter(self, name=None):
        """Property deleter."""
        if name is None:
            name = self.name
        self.deleternames.add(name)

    def test_empty_cls(self):
        """Test to add properties on an empty cls."""
        # add properties
        @addproperties(names=self.names)
        class Test(object):
            """Test class."""

        test = Test()

        for name in self.names:
            # assert properties are in Test
            prop = getattr(Test, name)
            self.assertTrue(isinstance(prop, property))
            # get protected attr name
            protectedattrname = _protectedattrname(name)
            # assert getter
            self.assertRaises(AttributeError, getattr, test, name)
            setattr(test, protectedattrname, self)
            value = getattr(test, name)
            self.assertIs(value, self)
            delattr(test, protectedattrname)
            # assert setter
            setattr(test, name, self)
            delattr(test, protectedattrname)
            # assert deleter
            self.assertRaises(AttributeError, delattr, test, name)
            setattr(test, protectedattrname, self)
            delattr(test, name)

    def test_not_empty_cls(self):
        """Test to add properties with existing properties."""
        class Test(object):
            """Test class."""

        for name in self.names:
            prop = property(
                fget=self.getter, fset=self.setter, fdel=self.deleter
            )
            setattr(Test, name, prop)

        addproperties(names=self.names)(Test)

        test = Test()

        for name in self.names:
            # assert existing getter
            self.name = name
            self.assertNotIn(name, self.getternames)
            value = getattr(test, name)
            self.assertIn(name, self.getternames)
            self.assertIs(value, self)
            # assert exising setter
            self.assertNotIn(name, self.setternames)
            setattr(test, name, value)
            self.assertIn(name, self.setternames)
            self.assertIs(self.setternames[name], self)
            # assert existing deleter
            self.assertNotIn(name, self.deleternames)
            delattr(test, name)
            self.assertIn(name, self.deleternames)

    def test_ab(self):
        """Test after/before getter/setter/deleter."""

        @addproperties(
            names=self.names,
            bfget=self.getter, afget=self.getter,
            bfset=self.setter, afset=self.setter,
            bfdel=self.deleter, afdel=self.deleter
        )
        class Test(object):
            """Test class."""

        test = Test()

        for name in self.names:
            # assert existing getter
            self.assertNotIn(name, self.getternames)
            self.assertRaises(AttributeError, getattr, test, name)
            # get protected attr name
            protectedattrname = _protectedattrname(name)
            setattr(test, protectedattrname, None)
            value = getattr(test, name)
            self.assertIn(name, self.getternames)
            self.assertIs(value, self)
            # assert exising setter
            self.assertNotIn(name, self.setternames)
            setattr(test, name, value)
            self.assertIn(name, self.setternames)
            self.assertIs(self.setternames[name], self)
            # assert existing deleter
            self.assertNotIn(name, self.deleternames)
            delattr(test, name)
            self.assertIn(name, self.deleternames)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Unit tests tools.
"""

from unittest import TestCase

from b3j0f.utils.version import PY2, PY26

from re import match

__all__ = ['UTCase']


def _subset(subset, superset):
    """True if subset is a subset of superset.

    :param dict subset: subset to compare.
    :param dict superset: superset to compare.
    :return: True iif all pairs (key, value) of subset are in superset.
    :rtype: bool
    """

    result = True
    for k in subset:
        result = k in superset and subset[k] == superset[k]
        if not result:
            break
    return result


class UTCase(TestCase):
    """Class which enrichs TestCase with python version compatibilities.
    """

    def __init__(self, *args, **kwargs):

        super(UTCase, self).__init__(*args, **kwargs)

    if PY2:  # python 3 compatibility

        if PY26:  # python 2.7 compatibility

            def assertIs(self, first, second, msg=None):
                return self.assertTrue(first is second, msg=msg)

            def assertIsNot(self, first, second, msg=None):
                return self.assertTrue(first is not second, msg=msg)

            def assertIn(self, first, second, msg=None):
                return self.assertTrue(first in second, msg=msg)

            def assertNotIn(self, first, second, msg=None):
                return self.assertTrue(first not in second, msg=msg)

            def assertIsNone(self, expr, msg=None):
                return self.assertTrue(expr is None, msg=msg)

            def assertIsNotNone(self, expr, msg=None):
                return self.assertFalse(expr is None, msg=msg)

            def assertIsInstance(self, obj, cls, msg=None):
                return self.assertTrue(isinstance(obj, cls), msg=msg)

            def assertNotIsInstance(self, obj, cls, msg=None):
                return self.assertTrue(not isinstance(obj, cls), msg=msg)

            def assertGreater(self, first, second, msg=None):
                return self.assertTrue(first > second, msg=msg)

            def assertGreaterEqual(self, first, second, msg=None):
                return self.assertTrue(first >= second, msg=msg)

            def assertLess(self, first, second, msg=None):
                self.assertTrue(first < second, msg=msg)

            def assertLessEqual(self, first, second, msg=None):
                return self.assertTrue(first <= second, msg=msg)

            def assertRegexpMatches(self, text, regexp, msg=None):
                return self.assertTrue(
                    match(regexp, text) if isinstance(regexp, str)
                    else regexp.search(text),
                    msg=msg
                )

            def assertNotRegexpMatches(self, text, regexp, msg=None):
                return self.assertIsNone(
                    match(regexp, text) if isinstance(regexp, str)
                    else regexp.search(text),
                    msg=msg
                )

            def assertItemsEqual(self, actual, expected, msg=None):
                return self.assertEqual(
                    sorted(actual), sorted(expected), msg=msg
                )

            def assertDictContainsSubset(self, expected, actual, msg=None):
                return self.assertTrue(_subset(expected, actual), msg=msg)

            def assertCountEqual(self, first, second, msg=None):
                return self.assertEqual(len(first), len(second), msg=msg)

        def assertRegex(self, text, regexp, msg=None):
            return self.assertRegexpMatches(text, regexp, msg)

        def assertNotRegex(self, text, regexp, msg=None):
            return self.assertNotRegexpMatches(text, regexp, msg)

    else:  # python 2 compatibility
        def assertRegexpMatches(self, *args, **kwargs):
            return self.assertRegex(*args, **kwargs)

        def assertNotRegexpMatches(self, *args, **kwargs):
            return self.assertNotRegex(*args, **kwargs)

        def assertItemsEqual(self, actual, expected, msg=None):
            return self.assertEqual(sorted(actual), sorted(expected), msg=msg)

        def assertDictContainsSubset(self, expected, actual, msg=None):
            return self.assertTrue(_subset(expected, actual), msg=msg)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Python reflection tools.
"""

from b3j0f.utils.version import PY2
from b3j0f.utils.iterable import ensureiterable

from inspect import isclass, isroutine, ismethod, getmodule

__all__ = ['base_elts', 'find_embedding', 'is_inherited']


def base_elts(elt, cls=None, depth=None):
    """Get bases elements of the input elt.

    - If elt is an instance, get class and all base classes.
    - If elt is a method, get all base methods.
    - If elt is a class, get all base classes.
    - In other case, get an empty list.

    :param elt: supposed inherited elt.
    :param cls: cls from where find attributes equal to elt. If None,
        it is found as much as possible. Required in python3 for function
        classes.
    :type cls: type or list
    :param int depth: search depth. If None (default), depth is maximal.
    :return: elt bases elements. if elt has not base elements, result is empty.
    :rtype: list
    """

    result = []

    elt_name = getattr(elt, '__name__', None)

    if elt_name is not None:

        cls = [] if cls is None else ensureiterable(cls)

        elt_is_class = False

        # if cls is None and elt is routine, it is possible to find the cls
        if not cls and isroutine(elt):

            if hasattr(elt, '__self__'):  # from the instance

                instance = elt.__self__  # get instance

                if instance is None and PY2:  # get base im_class if PY2
                    cls = list(elt.im_class.__bases__)

                else:  # use instance class
                    cls = [instance.__class__]

        # cls is elt if elt is a class
        elif isclass(elt):
            elt_is_class = True
            cls = list(elt.__bases__)

        if cls:  # if cls is not empty, find all base classes

            index_of_found_classes = 0  # get last visited class index
            visited_classes = set(cls)  # cache for visited classes
            len_classes = len(cls)

            if depth is None:  # if depth is None, get maximal value
                depth = -1  # set negative value

            while depth != 0 and index_of_found_classes != len_classes:
                len_classes = len(cls)

                for index in range(index_of_found_classes, len_classes):
                    _cls = cls[index]

                    for base_cls in _cls.__bases__:
                        if base_cls in visited_classes:
                            continue

                        else:
                            visited_classes.add(base_cls)
                            cls.append(base_cls)
                index_of_found_classes = len_classes
                depth -= 1

            if elt_is_class:
                # if cls is elt, result is classes minus first class
                result = cls

            elif isroutine(elt):

                # get an elt to compare with found element
                elt_to_compare = elt.__func__ if ismethod(elt) else elt

                for _cls in cls:  # for all classes
                    # get possible base elt
                    b_elt = getattr(_cls, elt_name, None)

                    if b_elt is not None:
                        # compare funcs
                        bec = b_elt.__func__ if ismethod(b_elt) else b_elt
                        # if matching, add to result
                        if bec is elt_to_compare:
                            result.append(b_elt)

    return result


def is_inherited(elt, cls=None):
    """True iif elt is inherited in a base class.

    :param elt: elt to check such as an inherited element.
    :param type cls: base cls where find the base elt.
    :return: true if elt is an inherited element.
    :rtype: bool
    """

    return base_elts(elt, cls=cls, depth=1)


def find_embedding(elt, embedding=None):
    """Try to get elt embedding elements.

    :param embedding: embedding element. Must have a module.

    :return: a list of [module [,class]*] embedding elements which define elt.
    :rtype: list
    """

    result = []  # result is empty in the worst case

    # start to get module
    module = getmodule(elt)

    if module is not None:  # if module exists

        visited = set()  # cache to avoid to visit twice same element

        if embedding is None:
            embedding = module

        # list of compounds elements which construct the path to elt
        compounds = [embedding]

        while compounds:  # while compounds elements exist
            # get last compound
            last_embedding = compounds[-1]
            # stop to iterate on compounds when last embedding is elt
            if last_embedding == elt:
                result = compounds  # result is compounds
                break

            else:
                # search among embedded elements
                for name in dir(last_embedding):
                    # get embedded element
                    embedded = getattr(last_embedding, name)

                    try:  # check if embedded has already been visited
                        if embedded not in visited:
                            visited.add(embedded)  # set it as visited

                        else:
                            continue

                    except TypeError:
                        pass

                    else:

                        # get embedded module
                        embedded_module = getmodule(embedded)
                        # and compare it with elt module
                        if embedded_module is module:
                            # add embedded to compounds
                            compounds.append(embedded)
                            # end the second loop
                            break

                else:
                    # remove last element if no coumpound element is found
                    compounds.pop(-1)

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase
from b3j0f.utils.chaining import Chaining, ListChaining


class ChainingTest(UTCase):
    """
    Test Chaining object.
    """

    def setUp(self):
        """
        Initialize object to embed into a Chaining.
        """

        self.content = None
        self.chaining = Chaining(self.content)

    def test_empty(self):
        """
        Test Chaining result without calls
        """

        results = self.chaining[:]

        self.assertFalse(results)

    def test_one(self):
        """
        Test one chaining of calls.
        """

        self.chaining.__hash__()
        results = self.chaining[:]

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0], hash(self.content))

        result = self.chaining[0]

        self.assertEqual(result, hash(self.content))

    def test_many(self):
        """
        Test several chaining of calls.
        """

        self.chaining.__hash__().__repr__()

        results = self.chaining[:]

        self.assertEqual(len(results), 2)
        self.assertEqual(results[0], hash(self.content))
        self.assertEqual(results[1], repr(self.content))
        self.assertEqual(results[-1], repr(self.content))

        first_result = self.chaining[0]
        self.assertEqual(first_result, hash(self.content))

        second_result = self.chaining[1]
        self.assertEqual(second_result, repr(self.content))

        last_result = self.chaining[-1]
        self.assertEqual(last_result, repr(self.content))

    def test_exception(self):
        """
        Test when an exception occured.
        """

        self.assertRaises(
            AttributeError,
            lambda: self.chaining.raiseexception
        )


class ListChainingTest(UTCase):
    """
    Test ListChaining object.
    """

    def setUp(self):
        """
        Initialize object to embed into a Chaining.
        """

        self.content = [None, '']
        self.chaining = ListChaining(*self.content)

    def test_empty(self):
        """
        Test Chaining result without calls
        """

        results = self.chaining[:]

        self.assertFalse(results)

    def test_one(self):
        """
        Test one chaining of calls.
        """

        self.chaining.__hash__()
        results = self.chaining[:]

        self.assertEqual(len(results), 1)
        self.assertEqual(
            results[0],
            [hash(self.content[0]), hash(self.content[1])]
        )

        result = self.chaining[0]

        self.assertEqual(result[0], hash(self.content[0]))
        self.assertEqual(result[1], hash(self.content[1]))

    def test_many(self):
        """
        Test several chaining of calls.
        """

        self.chaining.__hash__().__repr__()

        results = self.chaining[:]

        self.assertEqual(len(results), 2)
        self.assertEqual(
            results[0],
            [hash(self.content[0]), hash(self.content[1])]
        )
        self.assertEqual(
            results[1],
            [repr(self.content[0]), repr(self.content[1])]
        )
        self.assertEqual(
            results[-1],
            [repr(self.content[0]), repr(self.content[1])]
        )

        first_result = self.chaining[0]
        self.assertEqual(
            first_result,
            [hash(self.content[0]), hash(self.content[1])]
        )

        second_result = self.chaining[1]
        self.assertEqual(
            second_result,
            [repr(self.content[0]), repr(self.content[1])]
        )

        last_result = self.chaining[-1]
        self.assertEqual(
            last_result,
            [repr(self.content[0]), repr(self.content[1])]
        )

    def test_exception(self):
        """
        Test when an exception occured.
        """
        self.chaining.upper()

        results = self.chaining[:]

        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0][0], Exception)
        self.assertEqual(results[0][1], self.content[1].upper())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

__all__ = [
    'get_proxy', 'proxify_routine', 'proxify_elt', 'is_proxy', 'proxified_elt'
]

"""Module in charge of creating proxies like the design pattern ``proxy``.

A proxy is based on a callable element. It respects its signature but not the
implementation.
"""

from time import time

from functools import wraps

from types import MethodType, FunctionType

from opcode import opmap

from random import randint

from sys import maxsize

from inspect import (
    getmembers, isroutine, ismethod, getargspec, getfile, isbuiltin, isclass
)

from b3j0f.utils.version import PY2, PY3, basestring
from b3j0f.utils.path import lookup

# consts for interception loading
LOAD_GLOBAL = opmap['LOAD_GLOBAL']
LOAD_CONST = opmap['LOAD_CONST']

#: list of attributes to set after proxifying a function
WRAPPER_ASSIGNMENTS = ['__doc__', '__name__']
#: list of attributes to update after proxifying a function
WRAPPER_UPDATES = ['__dict__']
#: lambda function name
__LAMBDA_NAME__ = (lambda: None).__name__
#: proxy class name
__PROXY_CLASS__ = 'Proxy'
#: attribute name for proxified element
__PROXIFIED__ = '__proxified__'
#: instance method name for delegating proxy generation to the elt to proxify
__GETPROXY__ = '__getproxy__'


def proxify_elt(elt, bases=None, _dict=None, public=False):
    """Proxify input elt.

    :param elt: elt to proxify.
    :param bases: elt class base classes. If None, use elt type.
    :param dict _dict: specific elt class content to use.
    :param bool public: if True (default False), proxify only public members
        (where name starts with the character '_').
    :return: proxified element.
    :raises: TypeError if elt does not implement all routines of bases and
        _dict.
    """

    # ensure _dict is a dictionary
    proxy_dict = {} if _dict is None else _dict.copy()
    # set of proxified attribute names which are proxified during bases parsing
    # and avoid to proxify them twice during _dict parsing
    proxified_attribute_names = set()
    # ensure bases is a tuple of types
    if bases is None:
        bases = (elt if isclass(elt) else type(elt),)
    if isinstance(bases, basestring):
        bases = (lookup(bases),)
    elif isclass(bases):
        bases = (bases,)
    else:
        bases = tuple(bases)

    # fill proxy_dict with routines of bases
    for base in bases:
        # exclude object
        if base is object:
            continue
        for name, member in getmembers(base, isroutine):
            # check if name is public
            if public and not name.startswith('_'):
                continue
            eltmember = getattr(elt, name, None)
            if eltmember is None:
                raise TypeError(
                    'Wrong elt {0}. Must implement {1} ({2}) of {3}.'.
                    format(elt, name, member, base)
                )
            # proxify member if member is not a constructor
            if name not in ['__new__', '__init__']:
                # get routine from proxy_dict or eltmember
                routine = proxy_dict.get(name, eltmember)
                # exclude object methods
                if getattr(routine, '__objclass__', None) is not object:
                    # get routine proxy
                    routine_proxy = proxify_routine(routine)
                    if ismethod(routine_proxy):
                        routine_proxy = routine_proxy.__func__
                    # update proxy_dict
                    proxy_dict[name] = routine_proxy
                    # and save the proxified attribute flag
                    proxified_attribute_names.add(name)
    # proxify proxy_dict
    for name in proxy_dict:
        value = proxy_dict[name]
        if not hasattr(elt, name):
            raise TypeError(
                'Wrong elt {0}. Must implement {1} ({2}).'.format(
                    elt, name, value
                )
            )
        if isroutine(value):
            # if member has not already been proxified
            if name not in proxified_attribute_names:
                # proxify it
                value = proxify_routine(value)
            proxy_dict[name] = value
    # set default constructors if not present in proxy_dict
    if '__new__' not in proxy_dict:
        proxy_dict['__new__'] = object.__new__
    if '__init__' not in proxy_dict:
        proxy_dict['__init__'] = object.__init__
    # generate a new proxy class
    cls = type('Proxy', bases, proxy_dict)
    # instantiate proxy cls
    result = cls if isclass(elt) else cls()
    # bind elt to proxy
    setattr(result, __PROXIFIED__, elt)

    return result


def proxify_routine(routine, impl=None):
    """Proxify a routine with input impl.

    :param routine: routine to proxify.
    :param impl: new impl to use. If None, use routine.
    """

    # init impl
    impl = routine if impl is None else impl

    is_method = ismethod(routine)
    if is_method:
        function = routine.__func__
    else:
        function = routine

    # flag which indicates that the function is not a pure python function
    # and has to be wrapped
    wrap_function = not hasattr(function, '__code__')

    try:
        # get params from routine
        args, varargs, kwargs, _ = getargspec(function)
    except TypeError:
        # in case of error, wrap the function
        wrap_function = True

    if wrap_function:
        # if function is not pure python, create a generic one
        # with assignments
        assigned = []
        for wrapper_assignment in WRAPPER_ASSIGNMENTS:
            if hasattr(function, wrapper_assignment):
                assigned.append(wrapper_assignment)
        # and updates
        updated = []
        for wrapper_update in WRAPPER_UPDATES:
            if hasattr(function, wrapper_update):
                updated.append(wrapper_update)

        @wraps(function, assigned=assigned, updated=updated)
        def wrappedfunction(*args, **kwargs):
            """Default wrap function.
            """

        function = wrappedfunction
        # get params from function
        args, varargs, kwargs, _ = getargspec(function)

    name = function.__name__

    result = _compilecode(
        function=function, name=name, impl=impl,
        args=args, varargs=varargs, kwargs=kwargs
    )

    # set wrapping assignments
    for wrapper_assignment in WRAPPER_ASSIGNMENTS:
        try:
            value = getattr(function, wrapper_assignment)
        except AttributeError:
            pass
        else:
            setattr(result, wrapper_assignment, value)
    # set proxy module
    result.__module__ = proxify_routine.__module__
    # update wrapping updating
    for wrapper_update in WRAPPER_UPDATES:
        try:
            value = getattr(function, wrapper_update)
        except AttributeError:
            pass
        else:
            getattr(result, wrapper_update).update(value)

    # set proxyfied element on proxy
    setattr(result, __PROXIFIED__, routine)

    if is_method:  # create a new method
        args = [result, routine.__self__]
        if PY2:
            args.append(routine.im_class)
        result = MethodType(*args)

    return result


def _compilecode(function, name, impl, args, varargs, kwargs):
    """Get generated code.

    :return: function proxy generated code.
    :rtype: str
    """

    newcodestr, generatedname, impl_name = _generatecode(
        function=function, name=name, impl=impl,
        args=args, varargs=varargs, kwargs=kwargs
    )

    try:
        __file__ = getfile(function)
    except TypeError:
        __file__ = '<string>'

    # compile newcodestr
    code = compile(newcodestr, __file__, 'single')

    # define the code with the new function
    _globals = {}
    exec(code, _globals)

    # get new code
    newco = _globals[generatedname].__code__

    # get new consts list
    newconsts = list(newco.co_consts)

    if PY3:
        newcode = list(newco.co_code)
    else:
        newcode = [ord(co) for co in newco.co_code]

    consts_values = {impl_name: impl}

    # change LOAD_GLOBAL to LOAD_CONST
    index = 0
    newcodelen = len(newcode)
    while index < newcodelen:
        if newcode[index] == LOAD_GLOBAL:
            oparg = newcode[index + 1] + (newcode[index + 2] << 8)
            name = newco.co_names[oparg]
            if name in consts_values:
                const_value = consts_values[name]
                if const_value in newconsts:
                    pos = newconsts.index(const_value)
                else:
                    pos = len(newconsts)
                    newconsts.append(consts_values[name])
                newcode[index] = LOAD_CONST
                newcode[index + 1] = pos & 0xFF
                newcode[index + 2] = pos >> 8
        index += 1

    # get code string
    codestr = bytes(newcode) if PY3 else "".join([chr(co) for co in newcode])

    # get vargs
    vargs = [
        newco.co_argcount, newco.co_nlocals, newco.co_stacksize,
        newco.co_flags, codestr, tuple(newconsts), newco.co_names,
        newco.co_varnames, newco.co_filename, newco.co_name,
        newco.co_firstlineno, newco.co_lnotab,
        function.__code__.co_freevars,
        newco.co_cellvars
    ]
    if PY3:
        vargs.insert(1, newco.co_kwonlyargcount)

    # instanciate a new code object
    codeobj = type(newco)(*vargs)
    # instanciate a new function
    if function is None or isbuiltin(function):
        result = FunctionType(codeobj, {})
    else:
        result = type(function)(
            codeobj, function.__globals__, function.__name__,
            function.__defaults__, function.__closure__
        )

    return result


def _generatecode(function, name, impl, args, varargs, kwargs):

    code = ''

    # flag for lambda function
    islambda = __LAMBDA_NAME__ == name
    if islambda:
        generatedname = '_{0}'.format(int(time()))
    else:
        generatedname = name

    # get join method for reducing concatenation time execution
    join = "".join

    # default indentation
    indent = '    '

    if islambda:
        code = '{0} = lambda '.format(generatedname)
    else:
        code = 'def {0}('.format(generatedname)

    if args:
        code = join((code, '{0}'.format(args[0])))
    for arg in args[1:]:
        code = join((code, ', {0}'.format(arg)))

    if varargs is not None:
        if args:
            code = join((code, ', '))
        code = join((code, '*{0}'.format(varargs)))

    if kwargs is not None:
        if args or varargs is not None:
            code = join((code, ', '))
        code = join((code, '**{0}'.format(kwargs)))

    impl_name = '_{0}'.format(randint(0, maxsize))

    # insert impl call
    if islambda:
        code = join((code, ': {0}('.format(impl_name)))
    else:
        code = join(
            (
                code,
                '):\n{0}return {1}('.format(indent, impl_name)
            )
        )

    impl_args = args[1:] if ismethod(impl) else args
    if impl_args:
        code = join((code, '{0}'.format(impl_args[0])))
    for arg in impl_args[1:]:
        code = join((code, ', {0}'.format(arg)))

    if varargs is not None:
        if args:
            code = join((code, ', '))
        code = join((code, '*{0}'.format(varargs)))

    if kwargs is not None:
        if args or varargs is not None:
            code = join((code, ', '))
        code = join((code, '**{0}'.format(kwargs)))

    code = join((code, ')\n'))

    result = code, generatedname, impl_name

    return result


def get_proxy(elt, bases=None, _dict=None):
    """Get proxy from an elt.

    If elt implements the proxy generator method (named ``__getproxy__``), use
    it instead of using this module functions.

    :param elt: elt to proxify.
    :type elt: object or function/method
    :param bases: base types to enrich in the result cls if not None.
    :param _dict: class members to proxify if not None.
    """

    # try to find an instance proxy generator
    proxygenerator = getattr(elt, __GETPROXY__, None)

    # if a proxy generator is not found, use this module
    if proxygenerator is None:
        if isroutine(elt):
            result = proxify_routine(elt)

        else:  # in case of object, result is a Proxy
            result = proxify_elt(elt, bases=bases, _dict=_dict)

    else:  # otherwise, use the specific proxy generator
        result = proxygenerator()

    return result


def proxified_elt(proxy):
    """Get proxified element.

    :param proxy: proxy element from where get proxified element.
    :return: proxified element. None if proxy is not proxified.
    """

    if ismethod(proxy):
        proxy = proxy.__func__
    result = getattr(proxy, __PROXIFIED__, None)

    return result


def is_proxy(elt):
    """Return True if elt is a proxy.

    :param elt: elt to check such as a proxy.
    :return: True iif elt is a proxy.
    :rtype: bool
    """

    if ismethod(elt):
        elt = elt.__func__

    result = hasattr(elt, __PROXIFIED__)

    return result



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Tools for managing path resolution of python objects.
"""

# ensure str are unicodes
from __future__ import unicode_literals

from inspect import ismodule, currentframe

try:
    from importlib import import_module
except ImportError:
    import_module = __import__

from random import random

from b3j0f.utils.version import PY26

__all__ = ['clearcache', 'incache', 'lookup', 'getpath']

#: lookup cache
__LOOKUP_CACHE = {}


def clearcache(path=None):
    """Clear cache memory for input path.

    :param str path: element path to remove from cache. If None clear all cache

    :Example:

    >>> incache('b3j0f.utils')
    False
    >>> lookup('b3j0f.utils')
    >>> incache('b3j0f.utils')
    True
    >>> clearcache('b3j0f.path')
    >>> incache('b3j0f.utils')
    False
    >>> lookup('b3j0f.utils')
    >>> incache('b3j0f.utils')
    True
    >>> clearcache()
    >>> incache('b3j0f.utils')
    False
    """

    global __LOOKUP_CACHE

    if path is None:
        __LOOKUP_CACHE = {}
    else:
        __LOOKUP_CACHE.pop(path, None)


def incache(path):
    """Check if input path is in cache.

    :return: True if path is in cache
    :rtype: bool

    :Example:

    >>> incache('b3j0f.utils')
    False
    >>> lookup('b3j0f.utils')
    >>> incache('b3j0f.utils')
    True
    """

    return path in __LOOKUP_CACHE


def lookup(path, cache=True):
    """Get element reference from input element.

    :limitations: it does not resolve class methods or static values such as
        True, False, numbers, string and keywords.
    :param str path: full path to a python element.
    :param bool cache: if True (default), permits to reduce time complexity for
        lookup resolution in using cache memory to save resolved elements.
    :return: python object which is accessible through input path
        or raise an exception if the path is wrong.
    :rtype: object
    :raises ImportError: if path is wrong
    """

    result = None
    found = path and cache and path in __LOOKUP_CACHE

    if found:
        result = __LOOKUP_CACHE[path]

    elif path:

        # we generate a result in order to accept the result such as a None
        generated_result = random()
        result = generated_result

        components = path.split('.')
        index = 0
        components_len = len(components)

        module_name = components[0]

        # try to resolve an absolute path
        try:
            result = import_module(module_name)

        except ImportError:
            # resolve element globals or locals of the from previous frame
            previous_frame = currentframe().f_back

            if module_name in previous_frame.f_locals:
                result = previous_frame.f_locals[module_name]
            elif module_name in previous_frame.f_globals:
                result = previous_frame.f_globals[module_name]

        found = result is not generated_result

        if found:

            if components_len > 1:

                index = 1

                # try to import all sub-modules/packages
                try:  # check if name is defined from an external module
                    # find the right module
                    while index < components_len:
                        module_name = '{0}.{1}'.format(
                            module_name, components[index]
                        )
                        result = import_module(module_name)
                        index += 1

                except ImportError:
                    # path sub-module content
                    try:
                        if PY26:  # when __import__ is used
                            index = 1  # restart count of pathing
                        while index < components_len:
                            result = getattr(result, components[index])
                            index += 1

                    except AttributeError:
                        raise ImportError(
                            'Wrong path {0} at {1}'.format(
                                path, components[:index]
                            )
                        )
                else:  # in case of PY26
                    if PY26:
                        index = 1
                        while index < components_len:
                            result = getattr(result, components[index])
                            index += 1

            # save in cache if found
            if cache:
                __LOOKUP_CACHE[path] = result

    if not found:
        raise ImportError('Wrong path {0}'.format(path))

    return result


def getpath(element):
    """Get full path of a given element such as the opposite of the
    resolve_path behaviour.

    :param element: must be directly defined into a module or a package and has
        the attribute '__name__'.

    :return: element absolute path.
    :rtype: str

    :raises AttributeError: if element has not the attribute __name__.

    :Example:

    >>> getpath(getpath)
    b3j0f.utils.path.getpath
    """

    if not hasattr(element, '__name__'):
        raise AttributeError(
            'element {0} must have the attribute __name__'.format(element)
        )

    result = element.__name__ if ismodule(element) else \
        '{0}.{1}'.format(element.__module__, element.__name__)

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase
from b3j0f.utils.runtime import make_constants, bind_all

import random


class MakeConstants(UTCase):

    def setUp(self):
        self.output = []
        self.stoplist = ['range']

    def verbose(self, message):
        """
        Verbose function to apply when using make_constants
        """

        self.output.append(message)

    def sample(self):

        def sample(population, k):
            "Choose k unique random elements from a population sequence."
            if not isinstance(population, (list, tuple, str)):
                raise TypeError('Cannot handle type', type(population))
            n = len(population)
            if not 0 <= k <= n:
                raise ValueError("sample larger than population")
            result = [None] * k
            pool = list(population)
            for i in range(k):         # invariant:  non-selected at [0,n-i)
                j = int(random.random() * (n - i))
                result[i] = pool[j]
                pool[j] = pool[n - i - 1]
            return result

        return sample

    def _test_verbose(self):

        verbose_message = [
            "isinstance --> {0}".format(isinstance),
            "list --> {0}".format(list), "tuple --> {0}".format(tuple),
            "str --> {0}".format(str), "TypeError --> {0}".format(TypeError),
            "type --> {0}".format(type), "len --> {0}".format(len),
            "ValueError --> {0}".format(ValueError),
            "list --> {0}".format(list),
            #"range --> {0}".format(range),
            "int --> {0}".format(int),
            "random --> {0}".format(random),
            "new folded constant:{0}".format((list, tuple, str)),
            "new folded constant:{0}".format(random.random)
        ]

        self.assertEqual(self.output, verbose_message)

    def test_function(self):

        make_constants(
            verbose=self.verbose, stoplist=self.stoplist)(self.sample())

    def test_class(self):

        class A(object):

            pass

        A.a = self.sample()

        bind_all(A, verbose=self.verbose, stoplist=self.stoplist)

        self._test_verbose()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Module which provides variables and constants in order to ease developments
among several python versions and platforms.
"""

from __future__ import unicode_literals

from sys import version_info
from platform import python_implementation

__all__ = [
    'PY3', 'PY2', 'PY26', 'PY27',  # python versions
    'PYPY', 'CPYTHON', 'JYTHON', 'IRONPYTHON',  # python runtime types
    'basestring', 'getcallargs', 'OrderedDict'  # python2.7 objects
]

PY3 = version_info[0] == 3  #: python3
PY2 = version_info[0] == 2  #: python2
PY26 = PY2 and version_info[1] == 6  #: python2.6
PY27 = PY2 and version_info[1] == 7  #: python2.7
PYPY = python_implementation() == 'PyPy'  #: pypy
CPYTHON = python_implementation() == 'CPython'  #: cpython
JYTHON = python_implementation() == 'Jython'  #: jython
IRONPYTHON = python_implementation() == 'IronPython'  #: IronPython

if PY3:
    basestring = str
else:
    basestring = str, unicode

# add functions and types if py26 which exist in py27 and py3.x
if PY26:

    # add definition of ordereddict
    from ordereddict import OrderedDict

    # add functions and classes which come from
    from inspect import getargspec, ismethod
    from sys import getdefaultencoding

    # add definition of getcallargs
    def getcallargs(func, *positional, **named):
        """Get the mapping of arguments to values.

        A dict is returned, with keys the function argument names (including
        the names of the * and ** arguments, if any), and values the respective
        bound values from 'positional' and 'named'.
        """
        args, varargs, varkw, defaults = getargspec(func)
        f_name = func.__name__
        arg2value = {}

        # The following closures are basically because of tuple parameter
        # unpacking.
        assigned_tuple_params = []

        def assign(arg, value):
            if isinstance(arg, str):
                arg2value[arg] = value
            else:
                assigned_tuple_params.append(arg)
                value = iter(value)
                for i, subarg in enumerate(arg):
                    try:
                        subvalue = next(value)
                    except StopIteration:
                        raise ValueError('need more than %d %s to unpack' %
                                         (i, 'values' if i > 1 else 'value'))
                    assign(subarg, subvalue)
                try:
                    next(value)
                except StopIteration:
                    pass
                else:
                    raise ValueError('too many values to unpack')

        def is_assigned(arg):
            if isinstance(arg, str):
                return arg in arg2value
            return arg in assigned_tuple_params

        if ismethod(func) and func.im_self is not None:
            # implicit 'self' (or 'cls' for classmethods) argument
            positional = (func.im_self,) + positional
        num_pos = len(positional)
        num_total = num_pos + len(named)
        num_args = len(args)
        num_defaults = len(defaults) if defaults else 0
        for arg, value in zip(args, positional):
            assign(arg, value)
        if varargs:
            if num_pos > num_args:
                assign(varargs, positional[-(num_pos - num_args):])
            else:
                assign(varargs, ())
        elif 0 < num_args < num_pos:
            raise TypeError('%s() takes %s %d %s (%d given)' % (
                f_name, 'at most' if defaults else 'exactly', num_args,
                'arguments' if num_args > 1 else 'argument', num_total))
        elif num_args == 0 and num_total:
            if varkw:
                if num_pos:
                    # We should use num_pos, but Python also uses num_total:
                    raise TypeError('%s() takes exactly 0 arguments '
                                    '(%d given)' % (f_name, num_total))
            else:
                raise TypeError('%s() takes no arguments (%d given)' %
                                (f_name, num_total))
        for arg in args:
            if isinstance(arg, str) and arg in named:
                if is_assigned(arg):
                    raise TypeError("%s() got multiple values for keyword "
                                    "argument '%s'" % (f_name, arg))
                else:
                    assign(arg, named.pop(arg))
        if defaults:    # fill in any missing values with the defaults
            for arg, value in zip(args[-num_defaults:], defaults):
                if not is_assigned(arg):
                    assign(arg, value)
        if varkw:
            assign(varkw, named)
        elif named:
            unexpected = next(iter(named))
            if isinstance(unexpected, unicode):
                unexpected = unexpected.encode(getdefaultencoding(), 'replace')
            raise TypeError("%s() got an unexpected keyword argument '%s'" %
                            (f_name, unexpected))
        unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
        if unassigned:
            num_required = num_args - num_defaults
            raise TypeError('%s() takes %s %d %s (%d given)' % (
                f_name, 'at least' if defaults else 'exactly', num_required,
                'arguments' if num_required > 1 else 'argument', num_total))
        return arg2value

else:

    # add builtin objects from python2.7+
    from collections import OrderedDict
    from inspect import getcallargs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase

from re import compile as re_compile


class UTCaseTest(UTCase):

    def test_assertIs(self):
        self.assertIs(1, 1)

    def test_assertIsNot(self):
        self.assertIsNot(1, 2)

    def test_assertIn(self):
        self.assertIn(1, {1: 1})

    def test_assertNotIn(self):
        self.assertNotIn(1, {})

    def test_assertIsNone(self):
        self.assertIsNone(None)

    def test_assertIsNotNone(self):
        self.assertIsNotNone(1)

    def test_assertIsInstance(self):
        self.assertIsInstance(1, int)

    def test_assertNotIsInstance(self):
        self.assertNotIsInstance(None, type)

    def test_assertGreater(self):
        self.assertGreater(2, 1)

    def test_assertGreaterEqual(self):
        self.assertGreaterEqual(1, 1)
        self.assertGreaterEqual(2, 1)

    def test_assertLess(self):
        self.assertLess(1, 2)

    def test_assertLessEqual(self):
        self.assertLessEqual(1, 2)
        self.assertLessEqual(1, 1)

    def test_assertRegexpMatches(self):
        self.assertRegexpMatches('a', 'a')
        self.assertRegexpMatches('a', re_compile('a'))

    def test_assertRegex(self):
        self.assertRegex('a', 'a')
        self.assertRegex('a', re_compile('a'))

    def test_assertNotRegexpMatches(self):
        self.assertNotRegexpMatches('b', 'a')
        self.assertNotRegexpMatches('a', re_compile('b'))

    def test_assertNotRegex(self):
        self.assertNotRegex('a', 'b')
        self.assertNotRegex('a', re_compile('b'))

    def test_assertItemsEqual(self):
        self.assertItemsEqual([1, 2], [2, 1])

    def test_assertDictContainsSubset(self):
        self.assertDictContainsSubset({1: 2}, {1: 2, 3: 4})


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""b3j0f project building script.
"""

from setuptools import setup, find_packages

from os.path import abspath, dirname, join

# get long description from setup directory abspath
with open(join(dirname(abspath(__file__)), 'README.rst')) as f:
    DESC = f.read()

KEYWORDS = [
    'utils', 'chaining', 'iterable', 'tools', 'path', 'property', 'dynamic',
    'reflection', 'reflect', 'runtime', 'unittest', 'unit test', 'version',
    'common'
]

# add ordereddict if version for 2.6.x python version
DEPENDENCIES = ['ordereddict']

DESCRIPTION = 'Set of tools and utilities useful in python projects'

setup(
    name='b3j0f.utils',
    version='0.10.0',
    packages=find_packages(exclude=['test.*', '*.test.*']),
    author='b3j0f',
    author_email='jlabejof@yahoo.fr',
    install_requires=DEPENDENCIES,
    description=DESCRIPTION,
    long_description=DESC,
    include_package_data=True,
    url='https://github.com/b3j0f/utils/',
    license='MIT License',
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: French",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
        "Topic :: Software Development",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4"
    ],
    test_suite='b3j0f',
    keywords=KEYWORDS
)
